/* <one line to give the program's name and a brief idea of what it does.>
 *  Copyright (C) 2021 Javier S. Pedro

 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.

 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>

#include "pci.h"
#include "sbpro.h"
#include "viaac97.h"

/** VIA AC97 PCI vendor and product IDs */
#define VIA_PCI_VEND_ID 0x1106
#define VIA_PCI_PROD_ID 0x3058

enum {
	CFG_COMMAND         = 0x04, /* Word  */
	CFG_BAR0            = 0x10, /* DWord */
	CFG_BAR1            = 0x14, /* DWord */
	CFG_AC_INTF_STATUS  = 0x40,
	CFG_AC_LINK_CONTROL = 0x41,
	CFG_FUNCTION_ENABLE = 0x42,
	CFG_PNP_CONTROL     = 0x43,
	CFG_FM_NMI_CONTROL  = 0x48,
	CFG_GAME_PORT_ADDR  = 0x4A, /* Word */
};

/** Stores desired addresses for SBPro (i.e. contents of BLASTER env var) */
struct blaster {
	unsigned short addr;
	unsigned char irq;
	unsigned char dma;
	unsigned short midi;
	unsigned short joy;
} envblaster;

/** Detected PCI address for AC97 card */
static pcisel card;
/** Detected IO base of AC97 */
static unsigned short cardiobase;

/** Detect AC97 card and its IO base */
static void init_card(void)
{
	int err;
	uint16_t command;
	uint32_t bar;

	if ((err = pci_init_bios())) {
		printf("PCI bios is not present\n");
		exit(1);
	}
	
	if ((err = pci_find_device(&card, VIA_PCI_VEND_ID, VIA_PCI_PROD_ID, 0))) {
		printf("Could not find PCI AC'97 device\n");
		exit(1);
	}
	
	if ((err = pci_read_config_word(card, CFG_COMMAND, &command))
	    || !(command & 1)) {
		printf("Card is not initialized for IO access\n");
		exit(1);
	}
	
	if ((err = pci_read_config_dword(card, CFG_BAR0, &bar))) {
		printf("Could not read BAR0\n");
		exit(1);
	}
	
	cardiobase = bar & 0xFFFC;
}

/** Parse BLASTER env var given by str into a blaster struct */
static void parse_blaster(struct blaster *b, const char *str)
{
	char * s = (char*) str;

	/* Defaults */
	b->addr = 0x220;
	b->irq = 0x7;
	b->dma = 0x1;
	b->midi = 0;
	b->joy = 0;
	
	if (!s || !s[0]) return;
	
	while (s = strpbrk(s + 1, "AIDPJaidpj")) {
		switch (*s) {
			case 'A':
			case 'a':
				b->addr = strtol(s + 1, NULL, 16);
				break;
			case 'I':
			case 'i':
				b->irq = atoi(s + 1);
				break;
			case 'D':
			case 'd':
				b->dma = atoi(s + 1);
				break;
			case 'P':
			case 'p':
				b->midi = strtol(s + 1, NULL, 16);
				break;
			case 'J':
			case 'j':
				b->joy = strtol(s + 1, NULL, 16);
				break;
		}
	}
}

static void set_ac97_link(bool enable, bool sb, bool fm)
{
	uint8_t reg;
	pci_read_config_byte(card, CFG_AC_LINK_CONTROL, &reg);

	// Register 0x41, AC-Link Interface Control
	// Bit 0 SB PCM Data Output
	//     1 FM PCM Data Output
	//     2 SGD Read Channel PCM Data Output
	//     3 Variable-Sample-Rate On-Demand mode
	//     4 Serial Data Out
	//     5 Sync High
	//     6 Reset Low
	//     7 Enable

	reg &= ~0xC3;

	if (enable) {
		reg |= 0xC0; // Enable interface and deassert resert
		if (sb) reg |= 0x9; // Also enable VSR, though we never disable it
		if (fm) reg |= 0x2;
	}

	pci_write_config_byte(card, CFG_AC_LINK_CONTROL, reg);
}

static void set_sbpro(bool enable, bool fm, bool midi, bool gameport)
{
	uint8_t reg;
	pci_read_config_byte(card, CFG_FUNCTION_ENABLE, &reg);
	
	// Register 0x42, Function Enable
	// Bit 0 SoundBlaster Enable
	//     1 MIDI Enable
	//     2 FM Enable
	//     3 Game Port Enable
	//     4 Gate SoundBlaster PCM when FIFO empty
	
	reg &= ~0x1F;
	
	if (enable) {
		reg |= 0x15;
		if (fm)       reg |= 0x4;
		if (midi)     reg |= 0x2;
		if (gameport) reg |= 0x8;
	}
	
	pci_write_config_byte(card, CFG_FUNCTION_ENABLE, reg);
}

/** Configure legacy SB addreses as per contents of struct blaster */
static void config_sbpro_pnp(struct blaster *b)
{
	uint8_t config = 0;

	// Register 0x43, Plug and Play Control
	// Bits 7-6 IRQ : (00=IRQ5, 01=IRQ7, 10=IRQ9, 11=IRQ10)
	//      5-4 DMA : (00=DMA0, DMA1, DMA2, DMA3)
	//      3-2 MPU base : (00=300h, 310h, 320h, 330h)
	//      1-0 SB  base : (00=220h, 240h, 260h, 280h)
	
	switch (b->addr) {
	case 0x220:
		config |= 0;
		break;
	case 0x240:
		config |= 1;
		break;
	case 0x260:
		config |= 2;
		break;
	case 0x280:
		config |= 3;
		break;
	default:
		printf("Unsupported base address: A%x\n", b->addr);
		exit(1);
	}
	
	switch (b->midi) {
	case 0x300:
		config |= 0 << 2;
		break;
	case 0x310:
		config |= 1 << 2;
		break;
	case 0x320:
		config |= 2 << 2;
		break;
	case 0x330:
	case 0: // Leave at default if disabled
		config |= 3 << 2;
		break;
	default:
		printf("Unsupported MIDI port: P%x\n", b->midi);
		exit(1);
	}
	
	switch (b->dma) {
	case 0:
		config |= 0 << 4;
		break;
	case 1:
		config |= 1 << 4;
		break;
	case 2:
		config |= 2 << 4;
		break;
	case 3:
		config |= 3 << 4;
		break;
	default:
		printf("Unsupported DMA: I%d\n", b->irq);
		exit(1);
	}

	switch (b->irq) {
	case 5:
		config |= 0 << 6;
		break;
	case 7:
		config |= 1 << 6;
		break;
	case 9:
		config |= 2 << 6;
		break;
	case 10:
		config |= 3 << 6;
		break;
	default:
		printf("Unsupported IRQ: I%d\n", b->irq);
		exit(1);
	}
	
	pci_write_config_byte(card, CFG_PNP_CONTROL, config);
	
	if (b->joy) {
		pci_write_config_word(card, CFG_GAME_PORT_ADDR, b->joy);
	}
}

static void set_fm_nmi(bool trap, bool mix)
{
	uint8_t reg;
	pci_read_config_byte(card, CFG_FM_NMI_CONTROL, &reg);
	
	// Register 0x48, FM NMI Control
	// Bit 0 FM Trap Interrupt Disable
	//     1 FM SGD Data for SoundBlaster Mixing
	//     2 FM IRQ Select (1 SMI / 0 NMI)
	
	reg &= ~0x3;
	
	if (!trap) reg |= 1;
	if (mix)   reg |= 2;

	pci_write_config_byte(card, CFG_FM_NMI_CONTROL, reg);
}

static unsigned char get_master_volume(void)
{
	uint16_t value = viaac97_codec_read(cardiobase, CODEC_PRIMARY, VIAAC97_MIXER_REG_MASTER);
	int pvol;
	
	if (value & 0x8000) return 0; // Muted
	
	// Return R volume, bits 0-4
	pvol = value & 0x1F;
	pvol = (pvol * 100) / 0x1F;
	return 100 - pvol;
}

static void set_master_volume(unsigned char vol)
{
	int pvol = ((101 - vol) * 0x1F) / 100;
	uint16_t value = viaac97_codec_read(cardiobase, CODEC_PRIMARY, VIAAC97_MIXER_REG_MASTER);
	
	value &= ~0x8000; // Remove mute
	value &= ~0x1F;
	value &= ~(0x1F << 8);
	
	value |= (pvol & 0x1F);
	value |= (pvol & 0x1F) << 8;
	if (vol == 0) value |= 0x8000;
	
	viaac97_codec_write(cardiobase, CODEC_PRIMARY, VIAAC97_MIXER_REG_MASTER, value);
}

static unsigned char get_pcm_volume(void)
{
	uint16_t value = viaac97_codec_read(cardiobase, CODEC_PRIMARY, VIAAC97_MIXER_REG_PCM_OUT);
	int pvol;
	
	if (value & 0x8000) return 0; // Muted
	
	// Return R volume, bits 0-4
	pvol = (value & 0x1F);
	pvol = (pvol * 100) / 0x1F;
	return 100 - pvol;
}

static void set_pcm_volume(unsigned char vol)
{
	int pvol = ((101 - vol) * 0x1F) / 100;
	uint16_t value = viaac97_codec_read(cardiobase, CODEC_PRIMARY, VIAAC97_MIXER_REG_PCM_OUT);
	
	value &= ~0x8000; // Remove mute
	value &= ~0x1F;
	value &= ~(0x1F << 8);
	
	value |= (pvol & 0x1F);
	value |= (pvol & 0x1F) << 8;
	if (vol == 0) value |= 0x8000;
	
	viaac97_codec_write(cardiobase, CODEC_PRIMARY, VIAAC97_MIXER_REG_PCM_OUT, value);
}

static unsigned char get_pc_spkr_volume(void)
{
	uint16_t value = viaac97_codec_read(cardiobase, CODEC_PRIMARY, VIAAC97_MIXER_REG_PCBEEP);
	int pvol;
	
	if (value & 0x8000) return 0; // Muted

	// Bits 1-4
	pvol = (value >> 1) & 0xF;
	pvol = (pvol * 100) / 0xF;
	return 100 - pvol;
}

static void set_pc_spkr_volume(unsigned char vol)
{
	int pvol = ((101 - vol) * 15) / 100;
	uint16_t value = viaac97_codec_read(cardiobase, CODEC_PRIMARY, VIAAC97_MIXER_REG_PCBEEP);

	value &= ~0x8000; // Remove mute
	value &= ~0x1E;
	value |= (pvol & 0xF) << 1;
	if (vol == 0) value |= 0x8000;

	viaac97_codec_write(cardiobase, CODEC_PRIMARY, VIAAC97_MIXER_REG_PCBEEP, value);
}

static bool is_ac97_enabled()
{
	uint8_t reg;
	int err;

	if (!card) return false;
	
	err = pci_read_config_byte(card, CFG_AC_INTF_STATUS, &reg);
	if (err) return false;
	
	return reg & 1;
}

static bool is_sbpro_enabled()
{
	uint8_t reg;
	int err;

	if (!card) return false;
	
	err = pci_read_config_byte(card, CFG_FUNCTION_ENABLE, &reg);
	if (err) return false;
	
	return reg & 1;
}

static void dump_config_byte(unsigned char reg, const char *desc)
{
	uint8_t value;
	pci_read_config_byte(card, reg, &value);
	printf(" %02x %s: 0x%02x\n", reg, desc, value);
}

static void dump_config_word(unsigned char reg, const char *desc)
{
	uint16_t value;
	pci_read_config_word(card, reg, &value);
	printf(" %02x %s: 0x%04x\n", reg, desc, value);
}

static void dump_sbpro_mixer_reg(unsigned char reg, const char *desc)
{
	uint8_t value = sbpro_mixer_reg_get(envblaster.addr, reg);
	printf(" %02x %s: 0x%02x\n", reg, desc, value);
}

static void dump_ac97_reg(unsigned char reg, const char *desc)
{
	uint16_t value = viaac97_codec_read(cardiobase, CODEC_PRIMARY, reg);
	printf(" %02x %s: 0x%04x\n", reg, desc, value);
}

static void dump()
{
	printf("BLASTER: A%x I%d D%d P%x J%x\n",
	       envblaster.addr, envblaster.irq, envblaster.dma, envblaster.midi, envblaster.joy);
	
	printf("PCI card pcisel=0x%04x iobase=0x%04x\n", card, cardiobase);

	printf("Config space:\n");
	
	//dump_config_byte(CFG_AC_INTF_STATUS, "AC97 interface status");
	dump_config_byte(CFG_AC_LINK_CONTROL, "AC-Link interface control");
	dump_config_byte(CFG_FUNCTION_ENABLE, "Function enable");
	dump_config_byte(CFG_PNP_CONTROL, "PnP control");
	dump_config_byte(CFG_FM_NMI_CONTROL, "FM NMI control");
	dump_config_word(CFG_GAME_PORT_ADDR, "Game port base address");
	
	if (is_sbpro_enabled()) {
		printf("SBPro Mixer:\n");
		
		dump_sbpro_mixer_reg(SBPRO_MIXER_REG_VOICE, "Voice volume");
		dump_sbpro_mixer_reg(SBPRO_MIXER_REG_MIC, "Mic volume");
		dump_sbpro_mixer_reg(SBPRO_MIXER_REG_INPUT, "Input filter/Select");
		dump_sbpro_mixer_reg(SBPRO_MIXER_REG_OUTPUT, "Output filter/Stereo");
		dump_sbpro_mixer_reg(SBPRO_MIXER_REG_MASTER, "Master volume");
		dump_sbpro_mixer_reg(SBPRO_MIXER_REG_MIDI, "FM volume");
		dump_sbpro_mixer_reg(SBPRO_MIXER_REG_CD, "CD volume");
		dump_sbpro_mixer_reg(SBPRO_MIXER_REG_LINE, "Line volume");
	} else {
		printf("SBPro is disabled, not dumping mixer status\n");
	}
	
	if (is_ac97_enabled()) {
		printf("AC97 Mixer:\n");
		
		dump_ac97_reg(VIAAC97_MIXER_REG_MASTER, "Master Output Volume");
		//dump_ac97_reg(VIAAC97_MIXER_REG_AUX_OUT, "AUX Output Volume");
		dump_ac97_reg(VIAAC97_MIXER_REG_PCBEEP, "PC Beep Volume");
		dump_ac97_reg(VIAAC97_MIXER_REG_PCM_OUT, "PCM Output Volume");
	} else {
		printf("AC97 is disabled, not dumping mixer status\n");
	}
}

static void help()
{
	puts("VIAS <action> <args..>");
	puts("Supported actions:");
	puts("  enable|disable       enable or disable SBPro emulation support");
	puts("  dump                 dump registers from AC97 and SBPro");
	puts("  <mixer> [<newvol>]   AC97 volume control");
	puts("                         without <newvol>, prints current value");
	puts("                         Supported mixers: master, pcm, pcspkr");
	puts("  stereo               enables stereo output on SBPro");
	puts("  reset_mixer          resets both SBPro and AC97 mixer controls");
}

int main(int argc, const char *argv[]) 
{
	const char *action;
	
	if (argc >= 2) {
		action = argv[1];
	} else {
		action = "enable";
	}
	
	init_card();
	parse_blaster(&envblaster, getenv("BLASTER"));
	
	if (strcmp(action, "help") == 0 || strcmp(action, "?") == 0 || strcmp(action, "/?") == 0) {
		help();
		return 0;
	} else if (strcmp(action, "enable") == 0) {
		set_ac97_link(true, false, false);
		config_sbpro_pnp(&envblaster);
		set_sbpro(true, true, envblaster.midi, envblaster.joy);
		set_fm_nmi(false, true);
		set_ac97_link(true, true, true);

		// Sometimes, PCM output is muted on boot
		if (get_pcm_volume() == 0) {
			set_pcm_volume(75);
		}
	} else if (strcmp(action, "disable") == 0) {
		set_sbpro(false, false, false, false);
		set_fm_nmi(false, false);
		set_ac97_link(false, false, false); // Turn off AC codec entirely
	} else if (strcmp(action, "stereo") == 0) {
		uint8_t outputs = sbpro_mixer_reg_get(envblaster.addr, SBPRO_MIXER_REG_OUTPUT);
		outputs |= 0x22; // Enable "Stereo Switch" and "Output Filter"
		sbpro_mixer_reg_set(envblaster.addr, SBPRO_MIXER_REG_OUTPUT, outputs);
	} else if (strcmp(action, "master") == 0) {
		if (argc >= 3) {
			set_master_volume(atoi(argv[2]));
		}
		printf("Master: %d%%\n", get_master_volume());
	} else if (strcmp(action, "pcm") == 0) {
		if (argc >= 3) {
			set_pcm_volume(atoi(argv[2]));
		}
		printf("PCM Out: %d%%\n", get_pcm_volume());
	} else if (strcmp(action, "pcspkr") == 0) {
		if (argc >= 3) {
			set_pc_spkr_volume(atoi(argv[2]));
		}
		printf("PC Speaker: %d%%\n", get_pc_spkr_volume());
	} else if (strcmp(action, "reset_mixer") == 0) {
		sbpro_mixer_reset(envblaster.addr);
		viaac97_codec_write(cardiobase, CODEC_PRIMARY, 0, 1);
		set_pcm_volume(100);
	} else if (strcmp(action, "dump") == 0) {
		dump();
	} else {
		printf("Unknown action \"%s\"\n", action);
		return 1;
	}

	return 0;
}
